/*
 * gedit-print-job.h
 * This file is part of gedit
 *
 * Copyright (C) 2000-2001 Chema Celorio, Paolo Maggi
 * Copyright (C) 2002-2008 Paolo Maggi
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __GEDIT_PRINT_JOB_H__
#define __GEDIT_PRINT_JOB_H__

#include <gtk/gtk.h>
#include <gedit/gedit-view.h>

G_BEGIN_DECLS

#define GEDIT_TYPE_PRINT_JOB              (gedit_print_job_get_type())
#define GEDIT_PRINT_JOB(obj)              (G_TYPE_CHECK_INSTANCE_CAST((obj), GEDIT_TYPE_PRINT_JOB, GeditPrintJob))
#define GEDIT_PRINT_JOB_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST((klass), GEDIT_TYPE_PRINT_JOB, GeditPrintJobClass))
#define GEDIT_IS_PRINT_JOB(obj)           (G_TYPE_CHECK_INSTANCE_TYPE((obj), GEDIT_TYPE_PRINT_JOB))
#define GEDIT_IS_PRINT_JOB_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), GEDIT_TYPE_PRINT_JOB))
#define GEDIT_PRINT_JOB_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS((obj), GEDIT_TYPE_PRINT_JOB, GeditPrintJobClass))

typedef struct _GeditPrintJob		GeditPrintJob;
typedef struct _GeditPrintJobClass	GeditPrintJobClass;
typedef struct _GeditPrintJobPrivate	GeditPrintJobPrivate;

typedef enum
{
	GEDIT_PRINT_JOB_STATUS_PAGINATING,
	GEDIT_PRINT_JOB_STATUS_DRAWING
} GeditPrintJobStatus;

typedef enum
{
	GEDIT_PRINT_JOB_RESULT_OK,
	GEDIT_PRINT_JOB_RESULT_CANCEL,
	GEDIT_PRINT_JOB_RESULT_ERROR
} GeditPrintJobResult;

struct _GeditPrintJob
{
	GObject parent;

	/* <private> */
	GeditPrintJobPrivate *priv;
};

struct _GeditPrintJobClass
{
	GObjectClass parent_class;

        /* Signals */
	void (* printing)	(GeditPrintJob       *job,
				 GeditPrintJobStatus  status);

	void (* show_preview)	(GeditPrintJob       *job,
				 GtkWidget           *preview);

        void (* done)		(GeditPrintJob       *job,
        			 GeditPrintJobResult  result,
        			 const GError        *error);
};

GType			 gedit_print_job_get_type		(void) G_GNUC_CONST;

GeditPrintJob		*gedit_print_job_new			(GeditView                *view);

GtkPrintOperationResult	 gedit_print_job_print			(GeditPrintJob            *job,
								 GtkPrintOperationAction   action,
								 GtkPageSetup             *page_setup,
								 GtkPrintSettings         *settings,
								 GtkWindow                *parent,
								 GError                  **error);

void			 gedit_print_job_cancel			(GeditPrintJob            *job);

const gchar		*gedit_print_job_get_status_string	(GeditPrintJob            *job);

gdouble			 gedit_print_job_get_progress		(GeditPrintJob            *job);

GtkPrintSettings	*gedit_print_job_get_print_settings	(GeditPrintJob            *job);

GtkPageSetup		*gedit_print_job_get_page_setup		(GeditPrintJob            *job);

G_END_DECLS

#endif /* __GEDIT_PRINT_JOB_H__ */

/* ex:set ts=8 noet: */
